/*
 * Copyright (C) 2008 The Android Open Source Project
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.buzztime.tabletprovisioner.scene;

import com.android.ddmlib.IDevice;
import com.android.ddmlib.MultiLineReceiver;
import com.buzztime.tabletprovisioner.model.DevicePackageInfo;

import java.util.ArrayList;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

/**
 * Queries a device through 'adb shell dumpsys package p' to get the status of each package installed on it.
 */
public class PackagesLoader {
    public static DevicePackageInfo[] loadPackages( IDevice device ) {
        final ArrayList<DevicePackageInfo> devicePkgs = new ArrayList<DevicePackageInfo>();

        try {
            DumpsysPackageReceiver receiver = new DumpsysPackageReceiver(devicePkgs);
            device.executeShellCommand("dumpsys package p", receiver);

        } catch (Exception e) {
            // failed to query pm? force reinstall.
        }
        System.out.println("LoadPackagesTask::size="+devicePkgs.size());

        return devicePkgs.toArray(new DevicePackageInfo[devicePkgs.size()]);
    }


    /**
     * Receiver and parser for the "dumpsys package p" command.
     */
    private final static class DumpsysPackageReceiver extends MultiLineReceiver {
        StringBuilder sb = new StringBuilder();
        boolean inPackage = false;
        static final Pattern PACKAGE_START = // matches ^   Package [com.buzztime.bridgereceiver] (25fca9d8):
                Pattern.compile("Package\\s+\\[(\\S+)\\]\\s+\\(\\p{XDigit}+\\).*:");
        static final Pattern NAME_VALUE_PAIR = // matches name=value
                Pattern.compile("(\\S+)=(.*)");
        DevicePackageInfo buildPackage = null;

        ArrayList<DevicePackageInfo> devicePkgs;
        public DumpsysPackageReceiver(ArrayList<DevicePackageInfo> devicePkgs) {
            this.devicePkgs = devicePkgs;
        }

        @Override
        public void processNewLines(String[] lines) {
            for (String line : lines) {
//                System.out.println("line:"+line);
                Matcher matcher = PACKAGE_START.matcher(line);
                if (matcher.matches()) {
                    buildPackage = new DevicePackageInfo(matcher.group(1));
                    devicePkgs.add(buildPackage);
                    inPackage = true;
//                    System.out.println("DevicePackage: " + buildPackage.packageName);
                    continue;
                }
                if (inPackage) {
                    // break out each line into a property
                    Matcher nvpMatcher = NAME_VALUE_PAIR.matcher(line);
                    if (nvpMatcher.matches()){
                        String name = nvpMatcher.group(1);
                        String value = nvpMatcher.group(2);
                        switch(name){
                            case "versionCode": {  // versionCode=22 targetSdk=22
                                String[] parts = value.split("\\s+");
                                buildPackage.versionCode = Integer.decode(parts[0]);
                                if (value.contains("targetSdk")) {
                                    parts = parts[1].split("=");  // targetSdk=22
                                    buildPackage.targetSdk = Integer.decode(parts[1]);
                                }
                            }  break;
                            case "targetSdk": // targetSdk=22
                                buildPackage.targetSdk = Integer.decode(value );
                                break;
                            case "userId": // userId=10001 gids=[]
                                buildPackage.userId = Integer.decode(value.split("\\s+")[0] );
                                break;
                            case "codePath":
                                buildPackage.codePath = value;
                                break;
                            case "resourcePath":
                                buildPackage.resourcePath = value;
                                break;
                            case "versionName":
                                buildPackage.versionName = value;
                                break;
                            case "timeStamp":
                                buildPackage.timeStamp = value;
                                break;
                            case "lastUpdateTime":
                                buildPackage.lastUpdateTime = value;
                                break;
                            case "firstInstallTime":
                                buildPackage.firstInstallTime = value;
                                break;
                            default:
                        }
                    }

                }
            }
        }

        @Override
        public boolean isCancelled() {
            return false;
        }
    }

}
